/*
 * Author: Petter Reinholdtsen <pere@hungry.com>
 * date:   2001-01-20
 *
 * Alter ELF rpath information (insert, modify, remove).
 *
 * Based on source from Peeter Joot <peeterj@ca.ibm.com> and Geoffrey
 * Keating <geoffk@ozemail.com.au>.
 */

#ifdef HAVE_CONFIG_H
#  include "config.h"
#endif

#include <dlfcn.h>
#include <elf.h>
#include <fcntl.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#ifdef HAVE_GETOPT_H
#include <getopt.h>
#endif

typedef int (*killrpath_t)(const char *filename);
typedef int (*chrpath_t)(const char *filename, const char *newpath, int convert);

#ifdef HAVE_GETOPT_LONG
#  define GETOPT_LONG getopt_long

static struct option long_options[] =
{
  {"convert",   0, 0, 'c'},
  {"delete",    0, 0, 'd'},
  {"help",      0, 0, 'h'},
  {"keepgoing", 0, 0, 'k'},
  {"list",      0, 0, 'l'},
  {"replace",   1, 0, 'r'},
  {"version",   0, 0, 'v'},
  {0, 0, 0, 0}
};

#else /* not HAVE_GETOPT_LONG */
#  define GETOPT_LONG(argc,argv,optstr,lopts,lidx) getopt(argc,argv,optstr)
#endif /* not HAVE_GETOPT_LONG */

static void
usage(char *progname)
{
  printf("Usage: %s [-v|-d|-c|-r <path>] <program> [<program> ...]\n\n",
         progname);
  printf("   -v|--version                Display program version number\n");
  printf("   -d|--delete                 Delete current rpath/runpath setting\n");
#if defined(DT_RUNPATH)
  printf("   -c|--convert                Convert rpath to runpath\n");
#endif /* DT_RUNPATH */
  printf("   -r <path>|--replace <path>  Replace current rpath/runpath setting\n");
  printf("                               with the path given\n");
  printf("   -l|--list                   List the current rpath/runpath (default)\n");
  printf("   -k|--keepgoing              Do not fail on first error\n");
  printf("   -h|--help                   Show this usage information.\n");
#ifndef HAVE_GETOPT_LONG
  printf("\n *** The long options are not available on this platform");
#endif /* not HAVE_GETOPT_LONG */
#if !defined(DT_RUNPATH)
  printf("\n *** There is no support for runpath on this platform");
#endif /* DT_RUNPATH */
  printf("\n");
}

static unsigned
elf_class(const char *filename)
{
   Elf32_Ehdr ehdr;
   int fd;

   fd = open(filename, O_RDONLY);
   if (fd == -1)
     return 0;
   if (read(fd, &ehdr, sizeof(ehdr)) != sizeof(ehdr))
   {
     close(fd);
     return 0;
   }
   close(fd);
   if ((memcmp(ehdr.e_ident, ELFMAG, SELFMAG) != 0)
       || (ehdr.e_ident[EI_VERSION] != EV_CURRENT))
   {
     fprintf(stderr, "`%s' probably isn't an ELF file.\n", filename);
     return 0;
   }
   return ehdr.e_ident[EI_CLASS];
}

int
main(int argc, char * const argv[])
{
  int retval = 0;
  int convert = 0;      /* convert to given type */
  int remove = 0;       /* remove or not */
  int keep_going = 0;   /* Break on first error, or keep going? */
  char *newpath = NULL; /* insert this path */
  int opt;
#ifdef HAVE_GETOPT_LONG
  int option_index = 0;
#endif /* HAVE_GETOPT_LONG */
  void* dll[2];
  killrpath_t killrpath[2];
  chrpath_t chrpath[2];

  if (argc < 2)
    {
      usage(argv[0]);
      return 1;
    }

  do {
    opt = GETOPT_LONG(argc, argv, "cdhklr:v", long_options, &option_index);
    switch (opt)
      {
#if defined(DT_RUNPATH)
      case 'c':
        convert = 1;
        break;
#endif /* DT_RUNPATH */
      case 'd':
        remove = 1;
        break;
      case 'k':
        keep_going = 1;
        break;
      case 'r':
        newpath = optarg;
        break;
      case 'v':
        printf("%s version %s\n", PACKAGE, VERSION);
        exit(0);
        break;
      case 'l': /* This is the default action */
        newpath = NULL;
        break;
      case -1:
        break;
      default:
        printf("Invalid argument '%c'\n", opt);
	/* Fall through */
      case 'h':
        usage(argv[0]);
        exit(0);
        break;
      }
  } while (-1 != opt);

  dll[0] = dlopen("libchrpath32.so", RTLD_LAZY);
  killrpath[0] = (killrpath_t)dlsym(dll[0], "killrpath");
  chrpath[0] = (chrpath_t)dlsym(dll[0], "chrpath");

  dll[1] = dlopen("libchrpath64.so", RTLD_LAZY);
  killrpath[1] = (killrpath_t)dlsym(dll[1], "killrpath");
  chrpath[1] = (chrpath_t)dlsym(dll[1], "chrpath");

  while (optind < argc && (!retval || keep_going))
    {
      const char* program = argv[optind++];
      unsigned eclass = elf_class(program);
      if (!eclass)
      {
        retval = 1;
        continue;
      }
      if (remove)
        retval |= killrpath[eclass - ELFCLASS32](program);
      else
        /* list by default, replace if path is set */
        retval |= chrpath[eclass - ELFCLASS32](program, newpath, convert);
    }

  dlclose(dll[0]);
  dlclose(dll[1]);
  return retval;
}
